# Sample code: Create a memory-only collection, enroll multiple images belonging to different identities, then run 1 to N identification

import tfsdk
import os
options = tfsdk.ConfigurationOptions()
# Can set configuration options here

options.dbms = tfsdk.DATABASEMANAGEMENTSYSTEM.NONE # The data will not persist after the application terminates

# To have the collection data persist after application termination, use the SQLITE or POSTGRESQL backend option
# options.dbms = tfsdk.DATABASEMANAGEMENTSYSTEM.SQLITE
# options.dbms = tfsdk.DATABASEMANAGEMENTSYSTEM.POSTGRESQL

options.smallest_face_height = 40 # https://reference.trueface.ai/cpp/dev/latest/usage/general.html#_CPPv4N8Trueface20ConfigurationOptions18smallestFaceHeightE
# options.fd_mode = tfsdk.FACEDETECTIONMODE.VERSATILE
# options.fd_filter = tfsdk.FACEDETECTIONFILTER.BALANCED
options.fr_model = tfsdk.FACIALRECOGNITIONMODEL.LITE # TODO: Use the full model for improved accuracy

# TODO: If you have a NVIDIA gpu, then enable the enableGPU flag (you will require a GPU specific token for this).
# options.enable_GPU = True 

sdk = tfsdk.SDK(options)

# TODO: export your license token as TRUEFACE_TOKEN environment variable
is_valid = sdk.set_license(os.environ['TRUEFACE_TOKEN'])
if (is_valid == False):
    print("Invalid License Provided")
    print("Be sure to export your license token as TRUEFACE_TOKEN")
    quit()

# If we are using a database backend, need to first call create_database_connection
# Since we are using DATABASEMANAGEMENTSYSTEM.NONE this is not necessary

# ex. If using POSTGRESQL backend...
# res = sdk.create_database_connection("host=localhost port=5432 dbname=my_database user=postgres password=admin")
# if (res != tfsdk.ERRORCODE.NO_ERROR):
#   print("Unable to create database connection")
#   quit()

# ex. If using SQLITE backend...
# res = sdk.create_database_connection("my_database.db")
# if (res != tfsdk.ERRORCODE.NO_ERROR):
#   print("Unable to create database connection")
#   quit()

# Create a new collection
res = sdk.create_load_collection("my_collection")
if (res != tfsdk.ERRORCODE.NO_ERROR):
    print("Unable to create collection")
    quit()

# Some identities to populate into our collection
image_identities = [
    ("../images/armstrong_2.jpg", "Armstrong"),
    ("../images/armstrong_3.jpg", "Armstrong"), # Can add the same identity more than once
    ("../images/aldrin.jpg", "Aldrin")
]

for path, identity in image_identities:
    print("Processing image:", path, "with identity:", identity)
    # Generate a template for each image
    res = sdk.set_image(path)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("Unable to set image at path:", path)
        continue

    # Detect the largest face in the image
    found, faceBoxAndLandmarks = sdk.detect_largest_face()
    if found == False:
        print("No face detected in image:", path)
        continue

    # We want to only enroll high quality images into the database / collection
    # For more information, refer to the section titled "Selecting the Best Enrollment Images"
    # https://reference.trueface.ai/cpp/dev/latest/py/identification.html

    # Therefore, ensure that the face height is at least 100px
    faceHeight = faceBoxAndLandmarks.bottom_right.y - faceBoxAndLandmarks.top_left.y
    print("Face height:", faceHeight, "pixels")

    if faceHeight < 100:
        print("The face is too small in the image for a high quality enrollment, not enrolling")
        continue

    # Get the aligned chip so we can compute the image quality
    face = sdk.extract_aligned_face(faceBoxAndLandmarks)

    # Compute the image quality score
    res, quality = sdk.estimate_face_image_quality(face)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("There was an error computing the image quality")
        continue

    # Ensure the image quality is above a threshold
    print("Face quality:", quality)
    if quality < 0.8:
        print("The image quality is too poor for enrollment, not enrolling")
        continue

    # We can check the orientation of the head and ensure that it is facing forward
    # To see the effect of yaw and pitch on match score, refer to: https://reference.trueface.ai/cpp/dev/latest/py/face.html#tfsdk.SDK.estimate_head_orientation

    res, yaw, pitch, roll = sdk.estimate_head_orientation(faceBoxAndLandmarks)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("Unable to compute head orientation")
        continue

    yaw_deg = yaw * 180 / 3.14
    pitch_deg = yaw * 180 / 3.14

    if abs(yaw_deg) > 45:
        print("Enrollment image has too extreme a yaw, not enrolling")
        continue

    if abs(pitch_deg) > 30:
        print("Enrollment image has too extreme a pitch, not enrolling")
        continue

    # Finally ensure the user is not wearing a mask
    error_code, mask_label, score = sdk.detect_mask(faceBoxAndLandmarks)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("Unable to run mask detection")
        continue

    if (mask_label == tfsdk.MASKLABEL.MASK):
        print("Please choose a image with a mask for enrollment, not enrolling")
        continue

    # Now that we have confirmed the images are high quality, generate a template from that image
    res, faceprint = sdk.get_face_feature_vector(face)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("There was an error generating the faceprint")
        continue

    # Enroll the feature vector into the collection
    res, UUID = sdk.enroll_faceprint(faceprint, identity)
    if (res != tfsdk.ERRORCODE.NO_ERROR):
        print("Unable to enroll feature vector")
        continue

    # TODO: Can store the UUID for later use
    print("Success, enrolled template with UUID:", UUID)
    print("--------------------------------------------")


# Use a new image of Armstrong as the probe
res = sdk.set_image("../images/armstrong_1.jpg")
if (res != tfsdk.ERRORCODE.NO_ERROR):
    print("Unable to set image")
    quit()

# Extract the feature vector
# We do not need to check the quality of the probe face templates
# We mainly want to ensure that the enrollment templates are high quality, 
# This is less of a concern with probe templates 
res, v1, found_face = sdk.get_largest_face_feature_vector()
if (res != tfsdk.ERRORCODE.NO_ERROR or found_face != True):
    print("No face found in image, unable to generate feature vector")
    quit()

# Run identification query using threshold of 0.3
res, found, candidate = sdk.identify_top_candidate(v1, 0.3)

if found == False:
    print("Unable to find match above threshold")
else:
    print("Match found!")
    print("Identity:", candidate.identity)
    print("Match Probability:", candidate.match_probability)
    print("Similarity:", candidate.similarity_measure)




